import re
from urllib.parse import urljoin

import bs4
from bs4.element import Comment, NavigableString, Tag

_whitespace_to_space_regex = re.compile(r"[\n\t\r]+")
_reduce_spaces_regex = re.compile(r" {2,}")


def try_urljoin(base, url, allow_fragments=True):
    """attempts urljoin, on ValueError passes through url. Shortcuts http(s):// urls"""
    if url.startswith(("https://", "http://")):
        return url
    try:
        url = urljoin(base, url, allow_fragments=allow_fragments)
    except ValueError:
        pass
    return url


def get_attr(el, attr, check_name=None):
    """Get the attribute of an element if it exists.

    Args:
      el (bs4.element.Tag): a DOM element
      attr (string): the attribute to get
      check_name (string or list, optional): a list/tuple of strings or single
        string, that must match the element's tag name

    Returns:
      string: the attribute's value
    """
    if check_name is None:
        return el.get(attr)
    if isinstance(check_name, str) and el.name == check_name:
        return el.get(attr)
    if isinstance(check_name, (tuple, list)) and el.name in check_name:
        return el.get(attr)


def get_img_src_alt(img, dict_class, img_with_alt, base_url=""):
    """given a img element, returns both src and alt attributes as a list of tuples if alt exists, else returns the src as a string
    use for alt parsing with img
    """

    alt = get_attr(img, "alt", check_name="img")
    src = get_attr(img, "src", check_name="img")

    if src is not None:
        src = try_urljoin(base_url, src)

        if alt is None or not img_with_alt:
            return src
        else:
            return dict_class([("value", src), ("alt", alt)])


def get_children(node):
    """An iterator over the immediate children tags of this tag"""
    for child in node.children:
        if isinstance(child, bs4.Tag) and child.name != "template":
            yield child


def get_descendents(node):
    """An iterator over the all children tags (descendants) of this tag"""
    for desc in node.descendants:
        if isinstance(desc, bs4.Tag) and desc.name != "template":
            yield desc


def get_textContent(el, replace_img=False, img_to_src=True, base_url=""):
    """Get the text content of an element, replacing images by alt or src"""

    DROP_TAGS = ("script", "style", "template")
    PRE_TAGS = ("pre",)
    P_BREAK_BEFORE = 1
    P_BREAK_AFTER = 0
    PRE_BEFORE = 2
    PRE_AFTER = 3

    def text_collection(el, replace_img=False, img_to_src=True, base_url=""):
        # returns array of strings or integers

        items = []

        # drops the tags defined above and comments
        if el.name in DROP_TAGS or isinstance(el, Comment):
            items = []

        elif isinstance(el, NavigableString):
            value = el
            # replace \t \n \r by space
            value = _whitespace_to_space_regex.sub(" ", value)
            # replace multiple spaces with one space
            items = [_reduce_spaces_regex.sub(" ", value)]

        # don't do anything special for PRE-formatted tags defined above
        elif el.name in PRE_TAGS:
            items = [PRE_BEFORE, el.get_text(), PRE_AFTER]

        elif el.name == "img" and replace_img:
            value = el.get("alt")
            if value is None and img_to_src:
                value = el.get("src")
                if value is not None:
                    value = try_urljoin(base_url, value)

            if value is not None:
                items = [" ", value, " "]

        elif el.name == "br":
            items = ["\n"]

        else:
            for child in el.children:

                child_items = text_collection(child, replace_img, img_to_src, base_url)
                items.extend(child_items)

            if el.name == "p":
                items = [P_BREAK_BEFORE] + items
                items.append(P_BREAK_AFTER)

        return items

    results = [
        t for t in text_collection(el, replace_img, img_to_src, base_url) if t != ""
    ]

    if results:
        # remove <space> if it is first and last or if it is preceded by a <space> or <p> open/close
        length = len(results)
        for i in range(0, length):
            if results[i] == " " and (
                i == 0
                or i == length - 1
                or results[i - 1] == " "
                or results[i - 1] in (P_BREAK_BEFORE, P_BREAK_AFTER)
                or results[i + 1] == " "
                or results[i + 1] in (P_BREAK_BEFORE, P_BREAK_AFTER)
            ):
                results[i] = ""

    if results:
        # remove leading whitespace and <int> i.e. next lines
        while (
            isinstance(results[0], str) and (results[0] == "" or results[0].isspace())
        ) or results[0] in (P_BREAK_BEFORE, P_BREAK_AFTER):
            results.pop(0)
            if not results:
                break

    if results:
        # remove trailing whitespace and <int> i.e. next lines
        while (
            isinstance(results[-1], str)
            and (results[-1] == "" or results[-1].isspace())
        ) or results[-1] in (P_BREAK_BEFORE, P_BREAK_AFTER):
            results.pop(-1)
            if not results:
                break

    # trim leading and trailing non-<pre> whitespace
    if results:
        if isinstance(results[0], str):
            results[0] = results[0].lstrip()
        if isinstance(results[-1], str):
            results[-1] = results[-1].rstrip()

    # create final string by concatenating replacing consecutive sequence of <int> by largest value number of \n
    text = ""
    count = 0
    last = None
    for t in results:
        if t in (P_BREAK_BEFORE, P_BREAK_AFTER):
            count = max(t, count)
        elif t == PRE_BEFORE:
            text = text.rstrip(" ")
        elif not isinstance(t, int):
            if count or last == "\n":
                t = t.lstrip(" ")
            text = "".join([text, "\n" * count, t])
            count = 0
        last = t

    return text



import re

from ...date_util import DATETIME_RE, TIME_RE, normalize_dt
from ..dom_util import get_attr, get_img_src_alt, get_textContent, try_urljoin
from . import implied, vcp


def text(el, base_url=""):
    """Process p-* properties"""

    # handle value-class-pattern
    prop_value = vcp.text(el)
    if prop_value is not None:
        return prop_value

    prop_value = get_attr(el, "title", check_name=("abbr", "link"))
    if prop_value is None:
        prop_value = get_attr(el, "value", check_name=("data", "input"))
    if prop_value is None:
        prop_value = get_attr(el, "alt", check_name=("img", "area"))
    if prop_value is None:
        prop_value = get_textContent(el, replace_img=True, base_url=base_url)

    return prop_value


def url(el, dict_class, img_with_alt, base_url=""):
    """Process u-* properties"""

    prop_value = get_attr(el, "href", check_name=("a", "area", "link"))
    if prop_value is None:
        prop_value = get_img_src_alt(el, dict_class, img_with_alt, base_url)
        if prop_value is not None:
            return prop_value
    if prop_value is None:
        prop_value = get_attr(
            el, "src", check_name=("audio", "video", "source", "iframe")
        )
    if prop_value is None:
        prop_value = get_attr(el, "poster", check_name="video")
    if prop_value is None:
        prop_value = get_attr(el, "data", check_name="object")

    if prop_value is None:
        # handle value-class-pattern
        prop_value = vcp.text(el)

    if prop_value is None:
        prop_value = get_attr(el, "title", check_name="abbr")
    if prop_value is None:
        prop_value = get_attr(el, "value", check_name=("data", "input"))
    if prop_value is None:
        prop_value = get_textContent(el)

    return try_urljoin(base_url, prop_value)


def datetime(el, default_date=None):
    """Process dt-* properties

    Args:
      el (bs4.element.Tag): Tag containing the dt-value

    Returns:
      a tuple (string string): a tuple of two strings, (datetime, date)
    """

    # handle value-class-pattern
    prop_value = vcp.datetime(el, default_date)
    if prop_value is not None:
        return prop_value

    prop_value = get_attr(el, "datetime", check_name=("time", "ins", "del"))
    if prop_value is None:
        prop_value = get_attr(el, "title", check_name="abbr")
    if prop_value is None:
        prop_value = get_attr(el, "value", check_name=("data", "input"))
    if prop_value is None:
        prop_value = get_textContent(el)

    # if this is just a time, augment with default date
    match = re.match(TIME_RE + "$", prop_value)
    if match and default_date:
        prop_value = f"{default_date} {prop_value}"
        return normalize_dt(prop_value), default_date

    # otherwise, treat it as a full date
    match = re.match(DATETIME_RE + "$", prop_value)
    return (
        normalize_dt(prop_value, match=match),
        match and match.group("date"),
    )


def embedded(el, base_url=""):
    """Process e-* properties"""
    return {
        "html": el.decode_contents().strip(),  # secret bs4 method to get innerHTML
        "value": get_textContent(el, replace_img=True, base_url=base_url),
    }



"""Utility functions to deal with datetime strings."""

import datetime
import re

DATE_RE = r"(\d{4}-\d{2}-\d{2})|(\d{4}-\d{3})"
SEC_RE = r"(:(?P<second>\d{2})(\.\d+)?)"
RAWTIME_RE = rf"(?P<hour>\d{{1,2}})(:(?P<minute>\d{{2}}){SEC_RE}?)?"
AMPM_RE = r"am|pm|a\.m\.|p\.m\.|AM|PM|A\.M\.|P\.M\."
TIMEZONE_RE = r"Z|[+-]\d{1,2}:?\d{2}?"
TIME_RE = (
    rf"(?P<rawtime>{RAWTIME_RE})( ?(?P<ampm>{AMPM_RE}))?( ?(?P<tz>{TIMEZONE_RE}))?"
)
DATETIME_RE = rf"(?P<date>{DATE_RE})(?P<separator>[T ])(?P<time>{TIME_RE})"


def normalize_dt(dtstr, match=None):
    """Try to normalize a datetime string.
    1. Convert 12-hour time to 24-hour time

    pass match in if we have already calculated it to avoid rework
    """
    match = match or (dtstr and re.match(DATETIME_RE + "$", dtstr))
    if match:
        datestr = match.group("date")
        hourstr = match.group("hour")
        minutestr = match.group("minute") or "00"
        secondstr = match.group("second")
        ampmstr = match.group("ampm")
        separator = match.group("separator")

        # convert ordinal date YYYY-DDD to YYYY-MM-DD
        try:
            datestr = datetime.datetime.strptime(datestr, "%Y-%j").strftime("%Y-%m-%d")
        except ValueError:
            # datestr was not in YYYY-DDD format
            pass

        # 12 to 24 time conversion
        if ampmstr:
            hourstr = match.group("hour")
            hourint = int(hourstr)

            if (ampmstr.startswith("a") or ampmstr.startswith("A")) and hourint == 12:
                hourstr = "00"

            if (ampmstr.startswith("p") or ampmstr.startswith("P")) and hourint < 12:
                hourstr = hourint + 12

        dtstr = f"{datestr}{separator}{hourstr}:{minutestr}"

        if secondstr:
            dtstr += ":" + secondstr

        tzstr = match.group("tz")
        if tzstr:
            dtstr += tzstr
    return dtstr


def parse_dt(s):
    """The definition for microformats2 dt-* properties are fairly
    lenient.  This method converts an mf2 date string into either a
    datetime.date or datetime.datetime object. Datetimes will be naive
    unless a timezone is specified.

    :param str s: a mf2 string representation of a date or datetime
    :return: datetime.date or datetime.datetime
    :raises ValueError: if the string is not recognizable
    """

    if not s:
        return None

    s = re.sub(r"\s+", " ", s)
    date_re = r"(?P<year>\d{4,})-(?P<month>\d{1,2})-(?P<day>\d{1,2})"
    time_re = r"(?P<hour>\d{1,2}):(?P<minute>\d{2})(:(?P<second>\d{2})(\.(?P<microsecond>\d+))?)?"
    tz_re = r"(?P<tzz>Z)|(?P<tzsign>[+-])(?P<tzhour>\d{1,2}):?(?P<tzminute>\d{2})"
    dt_re = f"{date_re}((T| ){time_re} ?({tz_re})?)?$"

    m = re.match(dt_re, s)
    if not m:
        raise ValueError(f"unrecognized datetime {s}")

    year = m.group("year")
    month = m.group("month")
    day = m.group("day")

    hour = m.group("hour")

    if not hour:
        return datetime.date(int(year), int(month), int(day))

    minute = m.group("minute") or "00"
    second = m.group("second") or "00"

    if hour:
        dt = datetime.datetime(
            int(year), int(month), int(day), int(hour), int(minute), int(second)
        )
    if m.group("tzz"):
        dt = dt.replace(tzinfo=datetime.timezone.utc)
    else:
        tzsign = m.group("tzsign")
        tzhour = m.group("tzhour")
        tzminute = m.group("tzminute") or "00"

        if tzsign and tzhour:
            offset = datetime.timedelta(hours=int(tzhour), minutes=int(tzminute))
            if tzsign == "-":
                offset = -offset
            dt = dt.replace(
                tzinfo=datetime.timezone(offset, f"{tzsign}{tzhour}:{tzminute}")
            )

    return dt



import re

mf2_classes_re = re.compile("(p|e|u|dt|h)-((:?[a-z0-9]+-)?[a-z]+(:?-[a-z]+)*)$")
mf2_roots_re = re.compile("h-(:?[a-z0-9]+-)?[a-z]+(:?-[a-z]+)*$")
mf2_properties_re = re.compile("(p|e|u|dt)-(:?[a-z0-9]+-)?[a-z]+(:?-[a-z]+)*$")
mf2_e_properties_re = re.compile("e-(:?[a-z0-9]+-)?[a-z]+(:?-[a-z]+)*$")


def filter_classes(classes, regex=mf2_classes_re):
    """detect classes that are valid names for mf2, sort in dictionary by prefix"""

    types = {x: set() for x in ("u", "p", "dt", "e", "h")}
    for c in classes:
        match = regex.match(c)
        if match:
            if c[0] == "h":
                types["h"].add(c)
            else:
                types[match.group(1)].add(match.group(2))
    return types


def root(classes):
    return {c for c in classes if mf2_roots_re.match(c)}


def is_property_class(class_):
    return mf2_properties_re.match(class_)


def has_embedded_class(classes):
    return any(mf2_e_properties_re.match(c) for c in classes)



"""
Functions to parse the properties of elements according to the value class pattern

http://microformats.org/wiki/value-class-pattern

"""

import re

from ...date_util import (DATE_RE, DATETIME_RE, TIME_RE, TIMEZONE_RE,
                          normalize_dt)
from ..dom_util import get_children


def _get_vcp_value(el):
    if "value-title" in el.get("class", []):
        return el.get("title")
    return el.get_text()


def _get_vcp_children(el):
    return [
        c
        for c in get_children(el)
        if c.has_attr("class")
        and ("value" in c["class"] or "value-title" in c["class"])
    ]


def text(el):

    value_els = _get_vcp_children(el)
    if value_els:
        return "".join(_get_vcp_value(el) for el in value_els)


def datetime(el, default_date=None):

    value_els = _get_vcp_children(el)
    if value_els:
        date_parts = []
        for value_el in value_els:
            if "value-title" in value_el.get("class", []):
                title = el.get("title")
                if title:
                    date_parts.append(title.strip())
            elif value_el.name in ("img", "area"):
                alt = value_el.get("alt") or value_el.get_text()
                if alt:
                    date_parts.append(alt.strip())
            elif value_el.name == "data":
                val = value_el.get("value") or value_el.get_text()
                if val:
                    date_parts.append(val.strip())
            elif value_el.name == "abbr":
                title = value_el.get("title") or value_el.get_text()
                if title:
                    date_parts.append(title.strip())
            elif value_el.name in ("del", "ins", "time"):
                dt = value_el.get("datetime") or value_el.get_text()
                if dt:
                    date_parts.append(dt.strip())
            else:
                val = value_el.get_text()
                if val:
                    date_parts.append(val.strip())

        date_part = time_part = tz_part = None

        for part in date_parts:
            match = re.match(DATETIME_RE + "$", part)
            if match:
                # if it's a full datetime, then we're done
                date_part = match.group("date")
                return normalize_dt(part, match=match), date_part

            # only use first found value
            if re.match(TIME_RE + "$", part) and time_part is None:
                time_part = part
            elif re.match(DATE_RE + "$", part) and date_part is None:
                date_part = part
            elif re.match(TIMEZONE_RE + "$", part) and tz_part is None:
                tz_part = part

        # use default date
        if date_part is None:
            date_part = default_date

        if date_part and time_part:
            date_time_value = f"{date_part} {time_part}"
        else:
            date_time_value = date_part or time_part

        if tz_part:
            date_time_value += tz_part

        return normalize_dt(date_time_value), date_part



"""
Looks for classic microformats class names and augments them with
microformats2 names. Ported and adapted from php-mf2.

NOTE: functions in this module modify DOM elements. For this copies of the source tree are created, but BS4's copy
function doesn't create copy of all sub-elements: most notably, the values of the `attrs` dictionary are not copied,
and thus reference the same list objects as in the original tree. Thus, special care has to be taken when modifying the
tree so changes do not accidentally propagate.

a) adding new/removing children is safe

b) attributes (e.g. `class`) should only by changed by assigning a *copy* of the original value, not by modifying it in
place.

DO NOT:
child_classes = child.get('class', [])
child_classes.append('p-kittens')
child['class'] = child_classes

DO:
child_classes = child.get('class', [])[:] ###<------- COPY CREATED HERE
child_classes.append('p-kittens')
child['class'] = child_classes

"""

import copy
import json
import pathlib
from urllib.parse import unquote

import bs4

from .. import class_util
from ..dom_util import get_children

with (pathlib.Path(__file__).parent / "rules.json").open() as fp:
    classic_map = json.load(fp)


def _make_classes_rule(old_classes, new_classes):
    """Builds a rule for augmenting an mf1 class with its mf2
    equivalent(s).
    """

    def f(child, **kwargs):
        child_original = child.original or copy.copy(child)
        child_classes = child.get("class", [])[:]
        if all(cl in child_classes for cl in old_classes):
            child_classes.extend([cl for cl in new_classes if cl not in child_classes])
            child["class"] = child_classes

            # if any new class is e-* attach original to parse originally authored HTML
            if class_util.has_embedded_class(child_classes) and child.original is None:
                child.original = child_original

    return f


def _rel_tag_to_category_rule(child, html_parser, **kwargs):
    """rel=tag converts to p-category using a special transformation (the
    category becomes the tag href's last path segment). This rule adds a new data tag so that
    <a rel="tag" href="http://example.com/tags/cat"></a> gets replaced with
    <data class="p-category" value="cat"></data>
    """

    href = child.get("href", "")
    rels = child.get("rel", [])
    if "tag" in rels and href:
        segments = [seg for seg in href.split("/") if seg]
        if segments:
            if html_parser:
                soup = bs4.BeautifulSoup("", features=html_parser)
            else:
                soup = bs4.BeautifulSoup("")

            data = soup.new_tag("data")
            # this does not use what's given in the JSON
            # but that is not a problem currently
            # use mf1 class so it doesn't get removed later
            data["class"] = ["p-category"]
            data["value"] = unquote(segments[-1])
            child.insert_before(data)
            # remove tag from rels to avoid repeat
            # new list created, original not modifed -> safe on incomplete copy
            child["rel"] = [r for r in rels if r != "tag"]


def _make_rels_rule(old_rels, new_classes, html_parser):
    """Builds a rule for augmenting an mf1 rel with its mf2 class equivalent(s)."""

    # need to special case rel=tag as it operates differently

    def f(child, **kwargs):
        child_rels = child.get("rel", [])
        child_classes = child.get("class", [])[:]
        if all(r in child_rels for r in old_rels):
            if "tag" in old_rels:
                _rel_tag_to_category_rule(child, html_parser, **kwargs)
            else:
                child_classes.extend(
                    [cl for cl in new_classes if cl not in child_classes]
                )
                child["class"] = child_classes

    return f


def _get_rules(old_root, html_parser):
    """for given mf1 root get the rules as a list of functions to act on children"""

    class_rules = [
        _make_classes_rule(old_classes.split(), new_classes)
        for old_classes, new_classes in classic_map[old_root]
        .get("properties", {})
        .items()
    ]
    rel_rules = [
        _make_rels_rule(old_rels.split(), new_classes, html_parser)
        for old_rels, new_classes in classic_map[old_root].get("rels", {}).items()
    ]

    return class_rules + rel_rules


def root(classes):
    """get all backcompat root classnames"""
    return sorted(set([c for c in classes if c in classic_map]))


def apply_rules(el, html_parser):
    """add modern classnames for older mf1 classnames

    returns a copy of el and does not modify the original
    """

    el_copy = copy.copy(el)

    def apply_prop_rules_to_children(parent, rules):

        for child in get_children(parent):
            classes = child.get("class", [])[:]
            # find existing mf2 properties if any and delete them
            child["class"] = [
                cl for cl in classes if not class_util.is_property_class(cl)
            ]

            # apply rules to change mf1 to mf2
            for rule in rules:
                rule(child)

            # recurse if it's not a nested mf1 or mf2 root
            if not (class_util.root(classes) or root(classes)):
                apply_prop_rules_to_children(child, rules)

    # add mf2 root equivalent
    classes = el_copy.get("class", [])[:]
    old_roots = root(classes)
    for old_root in old_roots:
        new_roots = classic_map[old_root]["type"]
        classes.extend(new_roots)
    el_copy["class"] = classes

    # add mf2 prop equivalent to descendents and remove existing mf2 props
    rules = []
    for old_root in old_roots:
        rules.extend(_get_rules(old_root, html_parser))

    apply_prop_rules_to_children(el_copy, rules)

    return el_copy



"""
Utilities for interpreting mf2 data.

Microformats2 is a general way to mark up any HTML document with
classes and propeties. This module uses domain-specific assumptions
about the classes (specifically h-entry and h-event) to extract
certain interesting properties.

"""

import collections
import re
import string
import unicodedata
from urllib.parse import urljoin

import bs4
import easyuri

from . import date_util

URL_ATTRIBUTES = {
    "a": ["href"],
    "link": ["href"],
    "img": ["src"],
    "audio": ["src"],
    "video": ["src", "poster"],
    "source": ["src"],
}

# From https://indieweb.org/location#How_to_determine_the_location_of_a_microformat
LOCATION_PROPERTIES = frozenset(
    (
        "street-address",
        "extended-address",
        "post-office-box",
        "locality",
        "region",
        "postal-code",
        "country-name",
        "label",
        "latitude",
        "longitude",
        "altitude",
        "name",
    )
)


def get_url(parsed):
    """Given a property value that may be a list of simple URLs or complex
    h-* dicts (with a url property), extract a list of URLs. This is useful
    when parsing e.g., in-reply-to.

    Args:
      mf (string or dict): URL or h-cite-style dict

    Returns:
      list: a list of URLs
    """

    urls = []
    for item in parsed["items"]:
        if isinstance(item, str):
            urls.append(item)
        elif isinstance(item, dict) and any(
            x.startswith("h-") for x in item.get("type", [])
        ):
            urls.extend(item.get("properties", {}).get("url", []))

    return urls


def find_first_entry(parsed, types):
    """Find the first interesting h-* object in BFS-order

    :param dict parsed: a mf2py parsed dict
    :param list types: target types, e.g. ['h-entry', 'h-event']
    :return: an mf2py item that is one of `types`, or None
    """
    return next(_find_all_entries(parsed, types, False), None)


def find_all_entries(parsed, types, include_properties=False):
    """Find all h-* objects of a given type in BFS-order. Traverses the
    top-level items and their children and descendents. Includes property
    values (e.g. finding all h-cards would not find values of
    "p-author h-card") only if `include_properties` is True.

    :param dict parsed: a mf2py parsed dict
    :param list types: target types, e.g. ['h-entry', 'h-event']
    :param boolean include_properties: include properties in search of entries
    :return: all entries with any of the the target types
    """
    return list(_find_all_entries(parsed, types, include_properties))


def _find_all_entries(parsed, types, include_properties):
    queue = collections.deque(item for item in parsed["items"])
    while queue:
        item = queue.popleft()
        if any(h_class in item.get("type", []) for h_class in types):
            yield item
        queue.extend(item.get("children", []))
        if include_properties:
            queue.extend(
                prop
                for props in item.get("properties", {}).values()
                for prop in props
                if isinstance(prop, dict)
            )


def find_datetimes(parsed):
    """Find published, updated, start, and end dates.

    :param dict parsed: a mf2py parsed dict
    :return: a dictionary from property type to datetime or date
    """
    hentry = find_first_entry(parsed)
    result = {}

    if hentry:
        for prop in ("published", "updated", "start", "end"):
            date_strs = hentry["properties"].get(prop, [])
            result[prop] = date_util.parse_dt(" ".join(date_strs))


def get_plain_text(values, strip=True):
    """Get the first value in a list of values that we expect to be plain-text.
    If it is a dict, then return the value of "value".

    :param list values: a list of values
    :param boolean strip: true if we should strip the plaintext value
    :return: a string or None
    """
    if values:
        v = values[0]
        if isinstance(v, dict):
            v = v.get("value", "")
        if strip:
            v = v.strip()
        return v


def classify_comment(parsed, target_urls):
    """Find and categorize comments that reference any of a collection of
    target URLs. Looks for references of type reply, like, and repost.

    :param dict parsed: a mf2py parsed dict
    :param list target_urls: a collection of urls that represent the
      target post. this can include alternate or shortened URLs.
    :return: a list of applicable comment types ['like', 'reply', 'repost']
    """

    def process_references(objs, reftypes, result):
        for obj in objs:
            if isinstance(obj, dict):
                if any(
                    url in target_urls
                    for url in obj.get("properties", {}).get("url", [])
                ):
                    result += (r for r in reftypes if r not in result)
            elif obj in target_urls:
                result += (r for r in reftypes if r not in result)

    result = []
    hentry = find_first_entry(parsed, ["h-entry"])
    if hentry:
        reply_type = []
        if "rsvp" in hentry["properties"]:
            reply_type.append("rsvp")
        if "invitee" in hentry["properties"]:
            reply_type.append("invite")
        reply_type.append("reply")

        # TODO handle rel=in-reply-to
        for prop in ("in-reply-to", "reply-to", "reply"):
            process_references(hentry["properties"].get(prop, []), reply_type, result)

        for prop in ("like-of", "like"):
            process_references(hentry["properties"].get(prop, []), ("like",), result)

        for prop in ("repost-of", "repost"):
            process_references(hentry["properties"].get(prop, []), ("repost",), result)

    return result


def parse_author(obj):
    """Parse the value of a u-author property, can either be a compound
    h-card or a single name or url.

    :param object obj: the mf2 property value, either a dict or a string
    :result: a dict containing the author's name, photo, and url
    """
    result = {}
    if isinstance(obj, dict):
        names = obj["properties"].get("name")
        photos = obj["properties"].get("photo")
        urls = obj["properties"].get("url")
        if names:
            result["name"] = names[0]
        if photos:
            result["photo"] = photos[0]
        if urls:
            result["url"] = urls[0]
    elif obj:
        if obj.startswith("http://") or obj.startswith("https://"):
            result["url"] = obj
        else:
            result["name"] = obj
    return result


def find_author(parsed, source_url=None, hentry=None, fetch_mf2_func=None):
    """Use the authorship discovery algorithm
    https://indiewebcamp.com/authorship to determine an h-entry's
    author.

    :param dict parsed: an mf2py parsed dict.
    :param str source_url: the source of the parsed document.
    :param hentry dict: optional, the h-entry we're examining, if omitted,
        we'll just use the first one
    :param fetch_mf2_func callable: optional function that takes a URL
        and returns parsed mf2
    :return: a dict containing the author's name, photo, and url
    """

    def find_hentry_author(hentry):
        for obj in hentry["properties"].get("author", []):
            return parse_author(obj)

    def find_parent_hfeed_author(hentry):
        for hfeed in _find_all_entries(parsed, ["h-feed"], False):
            # find the h-entry's parent h-feed
            if hentry in hfeed.get("children", []):
                for obj in hfeed["properties"].get("author", []):
                    return parse_author(obj)

    if not hentry:
        hentry = find_first_entry(parsed, ["h-entry"])
        if not hentry:
            return None

    author_page = None

    # 3. if the h-entry has an author property, use that
    author = find_hentry_author(hentry)

    # 4. otherwise if the h-entry has a parent h-feed with author property,
    #    use that
    if not author:
        author = find_parent_hfeed_author(hentry)

    # 5. if an author property was found
    if author:
        # 5.2 otherwise if author property is an http(s) URL, let the
        #     author-page have that URL
        if list(author.keys()) == ["url"]:
            author_page = author["url"]
        # 5.1 if it has an h-card, use it, exit.
        # 5.3 otherwise use the author property as the author name,
        #     exit.
        else:
            return author

    # 6. if there is no author-page and the h-entry's page is a permalink page
    if not author_page:
        # 6.1 if the page has a rel-author link, let the author-page's
        #     URL be the href of the rel-author link
        rel_authors = parsed.get("rels", {}).get("author", [])
        if rel_authors:
            author_page = rel_authors[0]

    # 7. if there is an author-page URL
    if author_page:
        if not fetch_mf2_func:
            return {"url": author_page}

        # 7.1 get the author-page from that URL and parse it for microformats2
        parsed = fetch_mf2_func(author_page)
        hcards = find_all_entries(parsed, ["h-card"])

        # 7.2 if author-page has 1+ h-card with url == uid ==
        #     author-page's URL, then use first such h-card, exit.
        for hcard in hcards:
            hcard_url = get_plain_text(hcard["properties"].get("url"))
            hcard_uid = get_plain_text(hcard["properties"].get("uid"))
            if (
                hcard_url
                and hcard_uid
                and hcard_url == hcard_uid
                and hcard_url == author_page
            ):
                return parse_author(hcard)

        # 7.3 else if author-page has 1+ h-card with url property
        #     which matches the href of a rel-me link on the author-page
        #     (perhaps the same hyperlink element as the u-url, though not
        #     required to be), use first such h-card, exit.
        rel_mes = parsed.get("rels", {}).get("me", [])
        for hcard in hcards:
            hcard_url = get_plain_text(hcard["properties"].get("url"))
            if hcard_url and hcard_url in rel_mes:
                return parse_author(hcard)

        # 7.4 if the h-entry's page has 1+ h-card with url ==
        #     author-page URL, use first such h-card, exit.
        for hcard in hcards:
            hcard_url = get_plain_text(hcard["properties"].get("url"))
            if hcard_url and hcard_url == author_page:
                return parse_author(hcard)

        # 8. otherwise no deterministic author can be found.
        return None


def representative_hcard(parsed, source_url):
    """Find the representative h-card for a URL

    http://microformats.org/wiki/representative-h-card-parsing

    :param dict parsed: an mf2 parsed dict
    :param str source_url: the source of the parsed document.
    :return: the representative h-card if one is found
    """
    hcards = find_all_entries(parsed, ["h-card"], include_properties=True)
    # uid and url both match source_url
    for hcard in hcards:
        if source_url in hcard["properties"].get("uid", []) and source_url in hcard[
            "properties"
        ].get("url", []):
            return hcard
    # url that is also a rel=me
    for hcard in hcards:
        if any(
            url in parsed.get("rels", {}).get("me", [])
            for url in hcard["properties"].get("url", [])
        ):
            return hcard
    # single hcard with matching url
    found = None
    count = 0
    for hcard in hcards:
        if source_url in hcard["properties"].get("url", []):
            found = hcard
            count += 1
    if count == 1:
        return found


def convert_relative_paths_to_absolute(source_url, base_href, html):
    """Attempt to convert relative paths in foreign content
    to absolute based on the source url of the document. Useful for
    displaying images or links in reply contexts and comments.

    Gets list of tags/attributes from `URL_ATTRIBUTES`. Note that this
    function uses a regular expression to avoid adding a library
    dependency on a proper parser.

    :param str source_url: the source of the parsed document.
    :param str html: the text of the source document
    :return: the document with relative urls replaced with absolute ones
    """

    def do_convert(match):
        base_url = urljoin(source_url, base_href) if base_href else source_url
        return (
            match.string[match.start(0) : match.start(1)]
            + urljoin(base_url, match.group(1))
            + match.string[match.end(1) : match.end(0)]
        )

    if source_url:
        for tagname, attributes in URL_ATTRIBUTES.items():
            for attribute in attributes:
                pattern = re.compile(
                    rf"<{tagname}[^>]*?{attribute}\s*=\s*['\"](.*?)['\"]",
                    flags=re.DOTALL | re.MULTILINE | re.IGNORECASE,
                )
                html = pattern.sub(do_convert, html)

    return html


def is_name_a_title(name, content):
    """Determine whether the name property represents an explicit title.

    Typically when parsing an h-entry, we check whether p-name ==
    e-content (value). If they are non-equal, then p-name likely
    represents a title.

    However, occasionally we come across an h-entry that does not
    provide an explicit p-name. In this case, the name is
    automatically generated by converting the entire h-entry content
    to plain text. This definitely does not represent a title, and
    looks very bad when displayed as such.

    To handle this case, we broaden the equality check to see if
    content is a subset of name. We also strip out non-alphanumeric
    characters just to make the check a little more forgiving.

    :param str name: the p-name property that may represent a title
    :param str content: the plain-text version of an e-content property
    :return: True if the name likely represents a separate, explicit title
    """

    def normalize(s):
        if not isinstance(s, str):
            s = s.decode("utf-8")
        s = unicodedata.normalize("NFKD", s)
        s = s.lower()
        s = re.sub("[" + string.whitespace + string.punctuation + "]", "", s)
        return s

    if not content:
        return True
    if not name:
        return False
    return normalize(content) not in normalize(name)


def post_type_discovery(hentry):
    """Implementation of the post-type discovery algorithm
    defined here https://indiewebcamp.com/post-type-discovery#Algorithm

    :param dict hentry: mf2 item representing the entry to test

    :return: string, one of: 'org', 'person', 'event', 'rsvp',
                     'invite', 'reply', 'repost', 'like', 'photo',
                     'article', 'note', 'follow'

    """
    props = hentry.get("properties", {})
    if "h-card" in hentry.get("type", []):
        name = get_plain_text(props.get("name"))
        org = get_plain_text(props.get("org"))
        if name and org and name == org:
            return "org"
        return "person"

    if "h-event" in hentry.get("type", []):
        return "event"

    for prop, implied_type in [
        ("rsvp", "rsvp"),
        ("invitee", "invite"),
        ("in-reply-to", "reply"),
        ("repost-of", "repost"),
        ("like-of", "like"),
        ("follow-of", "follow"),
        ("photo", "photo"),
    ]:
        if props.get(prop) is not None:
            return implied_type
    # check name ~= content
    name = get_plain_text(props.get("name"))
    content = get_plain_text(props.get("content"))
    if not content:
        content = get_plain_text(props.get("summary"))
    if content and name and is_name_a_title(name, content):
        return "article"
    return "note"


def _interpret_common_properties(
    parsed,
    source_url,
    base_href,
    hentry,
    use_rel_syndication,
    want_json,
    fetch_mf2_func,
):
    result = {}
    props = hentry["properties"]

    for prop in ("url", "uid", "photo", "featured" "logo"):
        value = get_plain_text(props.get(prop))
        if value:
            result[prop] = value

    for prop in ("start", "end", "published", "updated", "deleted"):
        date_str = get_plain_text(props.get(prop))
        if date_str:
            if want_json:
                result[prop] = date_str
            else:
                result[prop + "-str"] = date_str
                try:
                    date = date_util.parse_dt(date_str)
                    if date:
                        result[prop] = date
                except ValueError:
                    raise ValueError(f"Failed to parse datetime {date_str}")

    author = find_author(parsed, source_url, hentry, fetch_mf2_func)
    if author:
        result["author"] = author

    content_prop = props.get("content")
    content_value = None
    if content_prop:
        if isinstance(content_prop[0], dict):
            content_html = content_prop[0].get("html", "").strip()
            content_value = content_prop[0].get("value", "").strip()
        else:
            content_value = content_html = content_prop[0]
        result["content"] = convert_relative_paths_to_absolute(
            source_url, base_href, content_html
        )
        result["content-plain"] = content_value

    summary_prop = props.get("summary")
    if summary_prop:
        if isinstance(summary_prop[0], dict):
            result["summary"] = summary_prop[0]["value"]
        else:
            result["summary"] = summary_prop[0]

    # Collect location objects, then follow this algorithm to consolidate their
    # properties:
    # https://indieweb.org/location#How_to_determine_the_location_of_a_microformat
    location_stack = [props]

    for prop in "location", "adr":
        vals = props.get(prop)
        if vals:
            if isinstance(vals[0], str):
                location_stack.append({"name": vals})
            else:
                location_stack.append(vals[0].get("properties", {}))

    geo = props.get("geo")
    if geo:
        if isinstance(geo[0], dict):
            location_stack.append(geo[0].get("properties", {}))
        else:
            if geo[0].startswith("geo:"):
                # a geo: URL. try to parse it. https://tools.ietf.org/html/rfc5870
                parts = geo[0][len("geo:") :].split(";")[0].split(",")
                if len(parts) >= 2:
                    location_stack.append(
                        {
                            "latitude": [parts[0]],
                            "longitude": [parts[1]],
                            "altitude": [parts[2]] if len(parts) >= 3 else [],
                        }
                    )

    for prop in LOCATION_PROPERTIES:
        for obj in location_stack:
            if obj and obj.get(prop) and not (obj == props and prop == "name"):
                result.setdefault("location", {})[prop] = obj[prop][0]

    if use_rel_syndication:
        result["syndication"] = list(
            set(
                parsed.get("rels", {}).get("syndication", [])
                + hentry["properties"].get("syndication", [])
            )
        )
    else:
        result["syndication"] = hentry["properties"].get("syndication", [])

    return result


def interpret_event(
    parsed,
    source_url,
    base_href=None,
    hevent=None,
    use_rel_syndication=True,
    want_json=False,
    fetch_mf2_func=None,
):
    """Given a document containing an h-event, return a dictionary::

        {
         'type': 'event',
         'url': the permalink url of the document (may be different than source_url),
         'start': datetime or date,
         'end': datetime or date,
         'name': plain-text event name,
         'content': body of event description (contains HTML)
        }

    :param dict parsed: the result of parsing a document containing mf2 markup
    :param str source_url: the URL of the parsed document, not currently used
    :param str base_href: (optional) the href value of the base tag
    :param dict hevent: (optional) the item in the above document representing
      the h-event. if provided, we can avoid a redundant call to
      find_first_entry
    :param boolean use_rel_syndication: (optional, default True) Whether
      to include rel=syndication in the list of syndication sources. Sometimes
      useful to set this to False when parsing h-feeds that erroneously include
      rel=syndication on each entry.
    :param boolean want_json: (optional, default false) if true, the result
      will be pure json with datetimes as strings instead of python objects
    :param callable fetch_mf2_func: (optional) function to fetch mf2 parsed
      output for a given URL.
    :return: a dict with some or all of the described properties
    """
    # find the h-event if it wasn't provided
    if not hevent:
        hevent = find_first_entry(parsed, ["h-event"])
        if not hevent:
            return {}

    result = _interpret_common_properties(
        parsed,
        source_url,
        base_href,
        hevent,
        use_rel_syndication,
        want_json,
        fetch_mf2_func,
    )
    result["type"] = "event"
    name_value = get_plain_text(hevent["properties"].get("name"))
    if name_value:
        result["name"] = name_value
    return result


def interpret_entry(
    parsed,
    source_url,
    base_href=None,
    hentry=None,
    use_rel_syndication=True,
    want_json=False,
    fetch_mf2_func=None,
):
    """Given a document containing an h-entry, return a dictionary::

        {
         'type': 'entry',
         'url': the permalink url of the document (may be different than source_url),
         'published': datetime or date,
         'updated': datetime or date,
         'name': title of the entry,
         'content': body of entry (contains HTML),
         'author': {
          'name': author name,
          'url': author url,
          'photo': author photo
         },
         'syndication': [
           'syndication url',
           ...
         ],
         'in-reply-to': [...],
         'like-of': [...],
         'repost-of': [...],
        }

    :param dict parsed: the result of parsing a document containing mf2 markup
    :param str source_url: the URL of the parsed document, used by the
      authorship algorithm
    :param str base_href: (optional) the href value of the base tag
    :param dict hentry: (optional) the item in the above document
      representing the h-entry. if provided, we can avoid a redundant
      call to find_first_entry
    :param boolean use_rel_syndication: (optional, default True) Whether
      to include rel=syndication in the list of syndication sources. Sometimes
      useful to set this to False when parsing h-feeds that erroneously include
      rel=syndication on each entry.
    :param boolean want_json: (optional, default False) if true, the result
      will be pure json with datetimes as strings instead of python objects
    :param callable fetch_mf2_func: (optional) function to fetch mf2 parsed
      output for a given URL.
    :return: a dict with some or all of the described properties
    """

    # find the h-entry if it wasn't provided
    if not hentry:
        hentry = find_first_entry(parsed, ["h-entry"])
        if not hentry:
            return {}

    result = _interpret_common_properties(
        parsed,
        source_url,
        base_href,
        hentry,
        use_rel_syndication,
        want_json,
        fetch_mf2_func,
    )
    if "h-cite" in hentry.get("type", []):
        result["type"] = "cite"
    else:
        result["type"] = "entry"

    title = get_plain_text(hentry["properties"].get("name"))
    if title and is_name_a_title(title, result.get("content-plain")):
        result["name"] = title

    for prop in (
        "in-reply-to",
        "like-of",
        "repost-of",
        "bookmark-of",
        "comment",
        "like",
        "repost",
    ):
        for url_val in hentry["properties"].get(prop, []):
            if isinstance(url_val, dict):
                result.setdefault(prop, []).append(
                    interpret(
                        parsed,
                        source_url,
                        base_href,
                        url_val,
                        use_rel_syndication=False,
                        want_json=want_json,
                        fetch_mf2_func=fetch_mf2_func,
                    )
                )
            else:
                result.setdefault(prop, []).append(
                    {
                        "url": url_val,
                    }
                )

    return result


def interpret_feed(
    parsed, source_url, base_href=None, hfeed=None, want_json=False, fetch_mf2_func=None
):
    """Interpret a source page as an h-feed or as an top-level collection
    of h-entries.

    :param dict parsed: the result of parsing a mf2 document
    :param str source_url: the URL of the source document (used for authorship
        discovery)
    :param str base_href: (optional) the href value of the base tag
    :param dict hfedd: (optional) the h-feed to be parsed. If provided,
        this will be used instead of the first h-feed on the page.
    :param callable fetch_mf2_func: (optional) function to fetch mf2 parsed
      output for a given URL.
    :return: a dict containing 'entries', a list of entries, and possibly other
        feed properties (like 'name').
    """
    result = {}
    # find the first feed if it wasn't provided
    if not hfeed:
        hfeed = find_first_entry(parsed, ["h-feed"])

    if hfeed:
        names = hfeed["properties"].get("name")
        if names:
            result["name"] = names[0]
        children = hfeed.get("children", [])
    # just use the top level 'items' as the feed children
    else:
        children = parsed.get("items", [])

    entries = []
    for child in children:
        entry = interpret(
            parsed,
            source_url,
            base_href,
            item=child,
            use_rel_syndication=False,
            want_json=want_json,
            fetch_mf2_func=fetch_mf2_func,
        )
        if entry:
            entries.append(entry)
    result["entries"] = entries
    return result


def interpret(
    parsed,
    source_url,
    base_href=None,
    item=None,
    use_rel_syndication=True,
    want_json=False,
    fetch_mf2_func=None,
):
    """Interpret a permalink of unknown type. Finds the first interesting
    h-* element, and delegates to :func:`interpret_entry` if it is an
    h-entry or :func:`interpret_event` for an h-event

    :param dict parsed: the result of parsing a mf2 document
    :param str source_url: the URL of the source document (used for authorship
      discovery)
    :param str base_href: (optional) the href value of the base tag
    :param dict item: (optional) the item to be parsed. If provided,
      this will be used instead of the first element on the page.
    :param boolean use_rel_syndication: (optional, default True) Whether
      to include rel=syndication in the list of syndication sources. Sometimes
      useful to set this to False when parsing h-feeds that erroneously include
      rel=syndication on each entry.
    :param boolean want_json: (optional, default False) If true, the result
      will be pure json with datetimes as strings instead of python objects
    :param callable fetch_mf2_func: (optional) function to fetch mf2 parsed
      output for a given URL.
    :return: a dict as described by interpret_entry or interpret_event, or None
    """
    if not item:
        item = find_first_entry(parsed, ["h-entry", "h-event"])

    if item:
        types = item.get("type", [])
        if "h-event" in types:
            return interpret_event(
                parsed,
                source_url,
                base_href=base_href,
                hevent=item,
                use_rel_syndication=use_rel_syndication,
                want_json=want_json,
                fetch_mf2_func=fetch_mf2_func,
            )
        elif "h-entry" in types or "h-cite" in types:
            return interpret_entry(
                parsed,
                source_url,
                base_href=base_href,
                hentry=item,
                use_rel_syndication=use_rel_syndication,
                want_json=want_json,
                fetch_mf2_func=fetch_mf2_func,
            )


def interpret_comment(
    parsed,
    source_url,
    target_urls,
    base_href=None,
    want_json=False,
    fetch_mf2_func=None,
):
    """Interpret received webmentions, and classify as like, reply, or
    repost (or a combination thereof). Returns a dict as described
    in :func:`interpret_entry`, with the additional fields::

        {
         'comment_type': a list of strings, zero or more of
                         'like', 'reply', or 'repost'
         'rsvp': a string containing the rsvp response (optional)
        }

    :param dict parsed: a parsed mf2 parsed document
    :param str source_url: the URL of the source document
    :param list target_urls: a collection containing the URL of the target\
      document, and any alternate URLs (e.g., shortened links) that should\
      be considered equivalent when looking for references
    :param str base_href: (optional) the href value of the base tag
    :param boolean want_json: (optional, default False) If true, the result
      will be pure json with datetimes as strings instead of python objects
    :param callable fetch_mf2_func: (optional) function to fetch mf2 parsed
      output for a given URL.
    :return: a dict as described above, or None
    """
    item = find_first_entry(parsed, ["h-entry"])
    if item:
        result = interpret_entry(
            parsed,
            source_url,
            base_href=base_href,
            hentry=item,
            want_json=want_json,
            fetch_mf2_func=fetch_mf2_func,
        )
        if result:
            result["comment_type"] = classify_comment(parsed, target_urls)
            rsvp = get_plain_text(item["properties"].get("rsvp"))
            if rsvp:
                result["rsvp"] = rsvp.lower()

            invitees = item["properties"].get("invitee")
            if invitees:
                result["invitees"] = [parse_author(inv) for inv in invitees]

        return result


# ===========================================================================


stable = {
    "adr": [
        "p-street-address",
        "p-extended-address",
        "p-post-office-box",
        "p-locality",
        "p-region",
        "p-postal-code",
        "p-country-name",
        "p-label",
        "p/u-geo",
        "p-latitude",
        "p-longitude",
        "p-altitude",
    ],
    "card": [
        "p-name",
        "p-honorific-prefix",
        "p-given-name",
        "p-additional-name",
        "p-family-name",
        "p-sort-string",
        "p-honorific-suffix",
        "p-nickname",
        "u-email",
        "u-logo",
        "u-photo",
        "u-url",
        "u-uid",
        "p-category",
        "p/h-adr",
        "p-post-office-box",
        "p-extended-address",
        "p-street-address",
        "p-locality",
        "p-region",
        "p-postal-code",
        "p-country-name",
        "p-label",
        "p/u/h-geo",
        "p-latitude",
        "p-longitude",
        "p-altitude",
        "p-tel",
        "p-note",
        "dt-bday",
        "u-key",
        "p-org",
        "p-job-title",
        "p-role",
        "u-impp",
        "p-sex",
        "p-gender-identity",
        "dt-anniversary",
    ],
    "entry": [
        "p-name",
        "p-summary",
        "e-content",
        "dt-published",
        "dt-updated",
        "p-author",
        "p-category",
        "u-url",
        "u-uid",
        "p-location",
        "u-syndication",
        "u-in-reply-to",
        "p-rsvp",
        "u-like-of",
        "u-repost-of",
    ],
    "event": [
        "p-name",
        "p-summary",
        "dt-start",
        "dt-end",
        "dt-duration",
        "e-content",
        "u-url",
        "p-category",
        "p-location(card/adr/geo)",
        "[p-attendee]",
    ],
    "feed": ["p-name", "p-author(card)", "u-url", "u-photo"],
    "geo": ["p-latitude", "p-longitude", "p-altitude"],
    "item": ["p-name", "u-url", "u-photo"],
    "product": [
        "p-name",
        "u-photo",
        "p-brand(card)",
        "p-category",
        "e-content",
        "u-url",
        "u-identifier",
        "p-review(review)",
        "p-price",
    ],
    "recipe": [
        "p-name",
        "p-ingredient",
        "p-yield",
        "e-instructions",
        "dt-duration",
        "u-photo",
        "p-summary",
        "p-author(card)",
        "dt-published",
        "p-nutrition",
        "p-category",
    ],
    "resume": [
        "p-name",
        "p-summary",
        "p-contact",
        "p-education(event+card)",
        "p-experience(event+card)",
        "p-skill",
        "p-affiliation",
    ],
    "review": [
        "p-name ",
        "p-item(card/event/adr/geo/product/item)",
        "p-author(card)",
        "dt-published",
        "p-rating",
        "p-best",
        "p-worst",
        "e-content",
        "p-category",
        "u-url",
    ],
    "review-aggregate": [
        "p-item(card/event/adr/geo/product/item)",
        "p-average",
        "p-best",
        "p-worst",
        "p-count",
        "p-votes",
        "p-name",
    ],
}
draft = {"app": ["p-name", "u-url", "u-logo", "u-photo"]}


def representative_card(mf2json: dict, source_url: str):
    """
    Return the representative card for given parsed document.

    http://microformats.org/wiki/representative-h-card-parsing

    """
    source = easyuri.parse(source_url).minimized
    cards = [
        card
        for card in _get_all_items(mf2json, ["h-card"], include_props=True)
        if (
            card["properties"].get("name", [""])[0]
            or card["properties"].get("nickname", [""])[0]
        )
    ]
    if match := _check_uid_and_url_match_source_url(cards, source):
        return match
    if match := _check_url_matches_rel_me(cards, mf2json):
        return match
    if match := _check_url_matches_source_url(cards, source):
        return match


def _check_uid_and_url_match_source_url(cards, source_url):  # FIXME same as below?
    """"""
    for card in cards:
        if source_url in _get_normalized_urls(
            card, "uid"
        ) and source_url in _get_normalized_urls(card, "url"):
            return card["properties"]


def _check_url_matches_rel_me(cards, parsed):
    """"""
    for card in cards:
        rel_mes = set()
        for rel_me in parsed.get("rels", {}).get("me", []):
            try:
                rel_me = easyuri.parse(rel_me)
            except ValueError:
                continue
            if isinstance(rel_me, (easyuri.HTTPURI, easyuri.HTTPSURI)):
                rel_mes.add(rel_me.minimized)
        if any(url in rel_mes for url in _get_normalized_urls(card, "url")):
            return card["properties"]


def _check_url_matches_source_url(cards, source_url):  # FIXME same as above?
    """"""
    found = []
    count = 0
    for card in cards:
        # if source_url in card['properties'].get('url', []):
        for card_url in _get_normalized_urls(card, "url"):
            if card_url.rstrip("/") == source_url:
                found.append(card)
                count += 1
    if count:
        return found[0]["properties"]


def representative_feed(mf2json: dict, source_url: str, source_dom=None):
    """
    Return the representative feed for given parsed document.

    https://indieweb.org/feed#How_To_Consume
    https://microformats.org/wiki/h-feed#Discovery

    """
    feed = {}
    try:
        feed["name"] = source_dom.select("title")[0].text
    except (AttributeError, IndexError):
        pass
    if author := representative_card(mf2json, source_url):
        feed["author"] = author
    items = []
    if first_feed := _get_first_item(mf2json, ["h-feed"]):
        if name := first_feed["properties"].get("name"):
            feed["name"] = [name]
        if authors := first_feed["properties"].get("author"):
            feed["author"] = []
            for author in authors:
                author["properties"]["type"] = author["type"]
                feed["author"].append(author["properties"])
        if children := first_feed["children"]:
            items = children
    else:
        items = _get_all_items(mf2json, ["h-entry", "h-event"])
    feed["items"] = []
    for item in items:
        item["properties"]["type"] = item["type"]
        feed["items"].append(item["properties"])
    if rel_next := mf2json["rels"].get("next"):
        feed["next"] = rel_next[0]
    if rel_prev := mf2json["rels"].get("prev"):
        feed["prev"] = rel_prev[0]
    return feed


def discover_post_type(properties):
    """
    Return the discovered post type.

    http://ptd.spec.indieweb.org/#x5-post-type-algorithm

    """
    type_specific_properties = {
        "rsvp": "rsvp",
        "repost-of": "repost",  # aka share
        "like-of": "like",  # aka favorite
        "in-reply-to": "reply",
        "listen-of": "listen",
        "bookmark-of": "bookmark",
        "checkin": "check-in",
        "video": "video",
        "audio": "audio",
        "photo": "photo",
        # TODO "checkin": "checkin",
        # TODO "bookmark-of": "bookmark",
        # TODO "follow-of": "follow",
        # TODO "weight": "weight",
    }
    for type_specific_property, post_type in type_specific_properties.items():
        if type_specific_property in properties:
            if (
                post_type in ("video", "audio", "photo")
                and "quotation-of" in properties
            ):
                return f"{post_type}/clip"
            return post_type
    content = ""
    try:
        content = _get_first_non_empty(properties["content"])
    except KeyError:
        try:
            content = _get_first_non_empty(properties["summary"])
        except KeyError:
            return "note"
    name = ""
    try:
        name = _get_first_non_empty(properties["name"])
    except KeyError:
        return "note"
    if name:
        try:
            content = dict(content)
        except ValueError:
            text_content = content
        else:
            text_content = bs4.BeautifulSoup(content["html"].strip()).text
        if not text_content.startswith(name):
            return "article"
    return "note"


def _get_first_item(mf2json: dict, item_type: set):
    """Return the first object(s) of given item_type(s) (eg. h-entry, h-event)."""
    return next(_yield_all_items(mf2json, item_type, False), None)


def _get_all_items(mf2json: dict, item_type: set, include_props=False):
    """Return all object(s) of given item_type(s) (eg. h-entry, h-event)."""
    return list(_yield_all_items(mf2json, item_type, include_props))


def _yield_all_items(mf2json: dict, item_type: set, include_props: bool):
    """
    Yield objects(s) of given item_type(s) in breadth first search.

    Traverses the top-level items and their children and descendents.
    Includes property values (e.g. finding all h-cards would not find
    values of "p-author h-card") only if `include_props` is True.

    """
    queue = collections.deque(item for item in mf2json["items"])
    while queue:
        item = queue.popleft()
        if any(h_class in item.get("type", []) for h_class in item_type):
            yield item
        queue.extend(item.get("children", []))
        if include_props:
            queue.extend(
                prop
                for props in item.get("properties", {}).values()
                for prop in props
                if isinstance(prop, dict)
            )


def _get_normalized_urls(card, prop):
    """Return a list of normalized URLs for an card's prop (uid/url)."""
    urls = []
    for url in card["properties"].get(prop, []):
        try:
            urls.append(easyuri.parse(url).minimized)
        except ValueError:
            pass
    return urls


def _get_first_non_empty(propval):
    """
    Return the first non-empty value in `propval`.

    If `propval` is not a list and non-empty, return it.

    """
    if not isinstance(propval, list):
        propval = [propval]
    for content in propval:
        if content:
            return content



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mf', 'mf.parser', 'mf.parser.backcompat', 'mf.parser.prop_util']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4.11.2,<5.0.0',
 'easyuri>=0.1.2',
 'html5lib>=1.1,<2.0',
 'requests>=2.28.2,<3.0.0',
 'txtint>=0.1.2']

entry_points = \
{'console_scripts': ['mf = mf:main']}

setup_kwargs = {
    'name': 'microformats',
    'version': '0.3.2',
    'description': 'tools for Microformats production, consumption and analysis',
    'long_description': '[microformats][0] are the simplest way to openly publish contacts, events,\nreviews, recipes, and other structured information on the web.\n\n>>> import mf\n>>> url = "https://alice.example"\n>>> doc = mf.parse(doc=f\'\'\'\n... <p class=h-card><a href={url}>Alice</a></p>\n... <ul class=h-feed>\n... <li class=h-entry>foo\n... <li class=h-entry>bar\n... </ul>\n... \'\'\', url=url)\n\n# >>> dict(doc)\n# >>> doc.json\n\n>>> card = doc["items"][0]\n>>> card["type"]\n[\'h-card\']\n>>> card["properties"]\n{\'name\': [\'Alice\'], \'url\': [\'https://alice.example\']}\n>>> feed = doc["items"][1]\n>>> feed["children"][0]["properties"]["name"]\n[\'foo\']\n\n>>> mf.util.representative_card(doc, url)\n{\'name\': [\'Alice\'], \'url\': [\'https://alice.example\']}\n>>> mf.util.representative_feed(doc, url)["items"][0]["name"]\n[\'foo\']\n\nBased upon [`mf2py`][1] and [`mf2util`][2].\n\n# TODO >>> doc.representative_card\n# TODO {\'name\': [\'Alice\'], \'url\': [\'https://alice.example\']}\n# TODO >>> doc.representative_feed["items"][0]["name"]\n# TODO [\'foo\']\n\n[0]: https://microformats.org/wiki/microformats\n[1]: https://github.com/microformats/mf2py\n[1]: https://github.com/kylewm/mf2util\n\n## Fork\n\n### mf2py\n\nStarted with commit:\nhttps://github.com/microformats/mf2py/commit/27f13087c1e0060381de9035280d56af5dad8649\n\nTom Morris <tom@tommorris.org> <https://tommorris.org>\nBarnaby Walters <https://waterpigs.co.uk>\nKartik Prabhu <me@kartikprabhu.com> <https://kartikprabhu.com>\nKyle Mahan <kyle@kylewm.com> <https://kylewm.com>\nKevin Marks <kevinmarks@gmail.com> <https://www.kevinmarks.com>\n\nCopyight (c) 2013, 2014 Tom Morris and contributors\n\nPermission is hereby granted, free of charge, to any person obtaining a copy\nof this software and associated documentation files (the "Software"), to deal\nin the Software without restriction, including without limitation the rights\nto use, copy, modify, merge, publish, distribute, sublicense, and/or sell\ncopies of the Software, and to permit persons to whom the Software is\nfurnished to do so, subject to the following conditions:\n\nThe above copyright notice and this permission notice shall be included in\nall copies or substantial portions of the Software.\n\nTHE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR\nIMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,\nFITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE\nAUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER\nLIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,\nOUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN\nTHE SOFTWARE.\n\n### mf2util\n\nStarted with commit:\nhttps://github.com/kylewm/mf2util/commit/b1acda62ea5b0d500dc5a6770b2c681825a01e41\n\nCopyright (c) 2014 Kyle Mahan\n\nRedistribution and use in source and binary forms, with or without\nmodification, are permitted provided that the following conditions are\nmet:\n\n1. Redistributions of source code must retain the above copyright\n   notice, this list of conditions and the following disclaimer.\n\n2. Redistributions in binary form must reproduce the above copyright\n   notice, this list of conditions and the following disclaimer in the\n   documentation and/or other materials provided with the\n   distribution.\n\nTHIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS\n"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT\nLIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR\nA PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT\nHOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,\nSPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT\nLIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,\nDATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY\nTHEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT\n(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE\nOF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.\n',
    'author': 'Angelo Gladding',
    'author_email': 'angelo@ragt.ag',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://ragt.ag/code/python-microformats',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



"""Microformats utilities."""

import collections
import importlib.metadata

import requests

from . import util
from .parser import Parser

__all__ = ["parse"]


def parse(doc=None, url=None, html_parser="html5lib", img_with_alt=False):
    """
    Return a dictionary containing the mf2json of the given HTML `doc`.

    You may provide a document, a URL or both. When both are provided
    the URL is used as the document's base href.

    You may specify an alternate `html_parser` with one of "html", "xml",
    "html5", "lxml", "html5lib" or "html.parser".

    Args:
      doc (file or string or BeautifulSoup doc): file handle, text of content
        to parse, or BeautifulSoup document. If None, it will be fetched from
        given url
      url (string): url of the file to be processed. Optionally extracted from
        base-element of given doc
    """
    url = str(url)
    metadata = importlib.metadata.metadata("microformats")
    useragent = f"{metadata['Home-page']} - version {metadata['Version']}"
    if doc is None:
        headers = {"User-Agent": useragent}
        try:
            response = requests.get(url, headers=headers)
        except requests.exceptions.SSLError:
            response = requests.get(url, headers=headers, verify=False)
        doc = response.text
        url = response.url
    mf2json = Parser(doc, url, html_parser, img_with_alt).to_dict()
    mf2json["debug"]["user-agent"] = useragent
    return mf2json



from .. import class_util
from ..dom_util import (get_attr, get_children, get_img_src_alt,
                        get_textContent, try_urljoin)


def name(el, base_url=""):
    """Find an implied name property

    Args:
      el (bs4.element.Tag): a DOM element

    Returns:
      string: the implied name value
    """

    def non_empty(val):
        """If alt or title is empty, we don't want to use it as the implied
        name"""
        return val is not None and val != ""

    # if image or area use alt text if not empty
    prop_value = get_attr(el, "alt", check_name=("img", "area"))
    if non_empty(prop_value):
        return prop_value

    # if abbreviation use the title if not empty
    prop_value = get_attr(el, "title", check_name="abbr")
    if non_empty(prop_value):
        return prop_value

    # find candidate child or grandchild
    poss_child = None
    children = list(get_children(el))
    if len(children) == 1:
        poss_child = children[0]

        # ignore if mf2 root
        if class_util.root(poss_child.get("class", [])):
            poss_child = None

        # if it is not img, area, abbr then find grandchild
        if poss_child and poss_child.name not in ("img", "area", "abbr"):
            grandchildren = list(get_children(poss_child))
            # if only one grandchild
            if len(grandchildren) == 1:
                poss_child = grandchildren[0]
                # if it is not img, area, abbr or is mf2 root then no possible child
                if poss_child.name not in ("img", "area", "abbr") or class_util.root(
                    poss_child.get("class", [])
                ):
                    poss_child = None

    # if a possible child was found
    if poss_child is not None:
        # use alt if possible child is img or area
        prop_value = get_attr(poss_child, "alt", check_name=("img", "area"))
        if non_empty(prop_value):
            return prop_value

        # use title if possible child is abbr
        prop_value = get_attr(poss_child, "title", check_name="abbr")
        if non_empty(prop_value):
            return prop_value

    # use text if all else fails
    # replace images with alt but not with src in implied name
    # proposal: https://github.com/microformats/microformats2-parsing/issues/35#issuecomment-393615508
    return get_textContent(el, replace_img=True, img_to_src=False, base_url=base_url)


def photo(el, dict_class, img_with_alt, base_url=""):
    """Find an implied photo property

    Args:
      el (bs4.element.Tag): a DOM element
      dict_class: a python class used as a dictionary (set by the Parser object)
      img_with_alt: a flag to enable experimental parsing of alt attribute with img (set by the Parser object)
      base_url (string): the base URL to use, to reconcile relative URLs

    Returns:
      string or dictionary: the implied photo value or implied photo as a dictionary with alt value
    """

    def get_photo_child(children):
        "take a list of children and finds a valid child for photo property"

        # if element has one image child use source if exists and img is
        # not root class
        poss_imgs = [c for c in children if c.name == "img"]
        if len(poss_imgs) == 1:
            poss_img = poss_imgs[0]
            if not class_util.root(poss_img.get("class", [])):
                return poss_img

        # if element has one object child use data if exists and object is
        # not root class
        poss_objs = [c for c in children if c.name == "object"]
        if len(poss_objs) == 1:
            poss_obj = poss_objs[0]
            if not class_util.root(poss_obj.get("class", [])):
                return poss_obj

    # if element is an img use source if exists
    prop_value = get_img_src_alt(el, dict_class, img_with_alt, base_url)
    if prop_value is not None:
        return prop_value

    # if element is an object use data if exists
    prop_value = get_attr(el, "data", check_name="object")
    if prop_value is not None:
        return prop_value

    # find candidate child or grandchild
    poss_child = None
    children = list(get_children(el))

    poss_child = get_photo_child(children)

    # if no possible child found then look for grandchild if only one child which is not not mf2 root
    if (
        poss_child is None
        and len(children) == 1
        and not class_util.root(children[0].get("class", []))
    ):

        grandchildren = list(get_children(children[0]))
        poss_child = get_photo_child(grandchildren)

    # if a possible child was found parse
    if poss_child is not None:
        # img get src
        prop_value = get_img_src_alt(poss_child, dict_class, img_with_alt, base_url)
        if prop_value is not None:
            return prop_value

        # object get data
        prop_value = get_attr(poss_child, "data", check_name="object")
        if prop_value is not None:
            return prop_value


def url(el, base_url=""):
    """Find an implied url property

    Args:
      el (bs4.element.Tag): a DOM element
      base_url (string): the base URL to use, to reconcile relative URLs

    Returns:
      string: the implied url value
    """

    def get_url_child(children):
        "take a list of children and finds a valid child for url property"

        # if element has one <a> child use if not root class
        poss_as = [c for c in children if c.name == "a"]
        if len(poss_as) == 1:
            poss_a = poss_as[0]
            if not class_util.root(poss_a.get("class", [])):
                return poss_a

        # if element has one area child use if not root class
        poss_areas = [c for c in children if c.name == "area"]
        if len(poss_areas) == 1:
            poss_area = poss_areas[0]
            if not class_util.root(poss_area.get("class", [])):
                return poss_area

    # if element is a <a> or area use its href if exists
    prop_value = get_attr(el, "href", check_name=("a", "area"))
    if prop_value is not None:  # an empty href is valid
        return try_urljoin(base_url, prop_value)

    # find candidate child or grandchild
    poss_child = None
    children = list(get_children(el))

    poss_child = get_url_child(children)

    # if no possible child found then look for grandchild if only one child which is not mf2 root
    if (
        poss_child is None
        and len(children) == 1
        and not class_util.root(children[0].get("class", []))
    ):

        grandchildren = list(get_children(children[0]))
        poss_child = get_url_child(grandchildren)

    # if a possible child was found parse
    if poss_child is not None:
        prop_value = get_attr(poss_child, "href", check_name=("a", "area"))
        if prop_value is not None:  # an empty href is valid
            return try_urljoin(base_url, prop_value)



"""
Microformats2 is a general way to mark up any HTML document with
classes and propeties. This library parses structured data from
a microformatted HTML document and returns a well-formed JSON
dictionary.

"""

import copy
import json
from urllib.parse import urlparse

from bs4 import BeautifulSoup, FeatureNotFound
from bs4.element import Tag

from . import backcompat, class_util
from .dom_util import get_attr, get_children, get_descendents, try_urljoin
from .prop_util import implied

__all__ = ["Parser"]


class Parser:
    """Object to parse a document for microformats and return them in
    appropriate formats.

    Args:
      doc (file or string or BeautifulSoup doc): file handle, text of content
        to parse, or BeautifulSoup document. If None, it will be fetched from
        given url
      url (string): url of the file to be processed. Optionally extracted from
        base-element of given doc
      html_parser (string): optional, select a specific HTML parser. Valid
        options from the BeautifulSoup documentation are:
        "html", "xml", "html5", "lxml", "html5lib", and "html.parser"
        defaults to "html5lib"

    """

    dict_class = dict

    def __init__(self, doc=None, url=None, html_parser="html5lib", img_with_alt=False):
        self.__url__ = None
        self.__doc__ = None
        self._preserve_doc = False
        self.__parsed__ = self.dict_class(
            [
                ("items", []),
                ("rels", self.dict_class()),
                ("rel-urls", self.dict_class()),
                (
                    "debug",
                    self.dict_class(),
                ),
            ]
        )
        self.__img_with_alt__ = img_with_alt

        # use default parser if none specified
        self.__html_parser__ = html_parser

        if url is not None:
            self.__url__ = url

        if doc is not None:

            if isinstance(doc, BeautifulSoup) or isinstance(doc, Tag):
                self.__doc__ = doc
                self._preserve_doc = True
            else:
                try:
                    # try the user-given html parser or default html5lib
                    self.__doc__ = BeautifulSoup(doc, features=self.__html_parser__)
                except FeatureNotFound:
                    # maybe raise a warning?
                    # else switch to default use
                    self.__doc__ = BeautifulSoup(doc)

        # update actual parser used
        # uses builder.NAME from BeautifulSoup
        if isinstance(self.__doc__, BeautifulSoup) and self.__doc__.builder is not None:
            self.__html_parser__ = self.__doc__.builder.NAME
        else:
            self.__html_parser__ = None

        # check for <base> tag
        if self.__doc__:

            poss_base = next(
                (el for el in get_descendents(self.__doc__) if el.name == "base"), None
            )
            if poss_base:
                poss_base_url = poss_base.get("href")  # try to get href
                if poss_base_url:
                    if urlparse(poss_base_url).netloc:
                        # base specifies an absolute path
                        self.__url__ = poss_base_url
                    elif self.__url__:
                        # base specifies a relative path
                        self.__url__ = try_urljoin(self.__url__, poss_base_url)

        if self.__doc__ is not None:
            # parse!
            self.parse()

    def parse(self):
        """Does the work of actually parsing the document. Done automatically
        on initialization.
        """
        self._default_date = None
        # _default_date exists to provide implementation for rules described
        # in legacy value-class-pattern. basically, if you have two dt-
        # properties and one does not have the full date, it can use the
        # existing date as a template.
        # see value-class-pattern#microformats2_parsers on wiki.
        # see also the implied_relative_datetimes testcase.

        def handle_microformat(
            root_class_names,
            el,
            value_property=None,
            simple_value=None,
            backcompat_mode=False,
        ):
            """Handles a (possibly nested) microformat, i.e. h-*"""
            properties = self.dict_class()
            children = []
            self._default_date = None
            # for processing implied properties: collects if property types (p, e, u, d(t)) or children (h) have been processed
            parsed_types_aggregation = set()

            if backcompat_mode:
                el = backcompat.apply_rules(el, self.__html_parser__)
                root_class_names = class_util.root(el.get("class", []))

            # parse for properties and children
            for child in get_children(el):
                (
                    child_props,
                    child_children,
                    child_parsed_types_aggregation,
                ) = parse_props(child)
                for key, new_value in child_props.items():
                    prop_value = properties.get(key, [])
                    prop_value.extend(new_value)
                    properties[key] = prop_value
                children.extend(child_children)
                parsed_types_aggregation.update(child_parsed_types_aggregation)

            # complex h-* objects can take their "value" from the
            # first explicit property ("name" for p-* or "url" for u-*)
            if value_property and value_property in properties:
                simple_value = properties[value_property][0]

            # if some properties not already found find in implied ways unless in backcompat mode
            if not backcompat_mode:
                # stop implied name if any p-*, e-*, h-* is already found
                if "name" not in properties and parsed_types_aggregation.isdisjoint(
                    "peh"
                ):
                    properties["name"] = [implied.name(el, base_url=self.__url__)]

                if "photo" not in properties and parsed_types_aggregation.isdisjoint(
                    "uh"
                ):
                    x = implied.photo(
                        el,
                        self.dict_class,
                        self.__img_with_alt__,
                        base_url=self.__url__,
                    )
                    if x is not None:
                        properties["photo"] = [x]

                # stop implied url if any u-* or h-* is already found
                if "url" not in properties and parsed_types_aggregation.isdisjoint(
                    "uh"
                ):
                    x = implied.url(el, base_url=self.__url__)
                    if x is not None:
                        properties["url"] = [x]

            # build microformat with type and properties
            microformat = self.dict_class(
                [
                    ("type", [class_name for class_name in sorted(root_class_names)]),
                    ("properties", properties),
                ]
            )
            if el.name == "area":
                shape = get_attr(el, "shape")
                if shape is not None:
                    microformat["shape"] = shape

                coords = get_attr(el, "coords")
                if coords is not None:
                    microformat["coords"] = coords

            # insert children if any
            if children:
                microformat["children"] = children

            Id = get_attr(el, "id")
            if Id:
                microformat["id"] = Id

            # simple value is the parsed property value if it were not
            # an h-* class
            if simple_value is not None:
                if isinstance(simple_value, dict):
                    # for e-* properties, the simple value will be
                    # {"html":..., "value":...}  which we should fold
                    # into the microformat object
                    # details: https://github.com/microformats/mf2py/issues/35
                    microformat.update(simple_value)
                else:
                    microformat["value"] = simple_value

            return microformat

        def parse_props(el):
            """Parse the properties from a single element"""
            props = self.dict_class()
            children = []
            # for processing implied properties: collects if property types (p, e, u, d(t)) or children (h) have been processed
            parsed_types_aggregation = set()

            classes = el.get("class", [])
            filtered_classes = class_util.filter_classes(classes)
            # Is this element a microformat2 root?
            root_class_names = filtered_classes["h"]
            backcompat_mode = False

            # Is this element a microformat1 root?
            if not root_class_names:
                root_class_names = backcompat.root(classes)
                backcompat_mode = True

            if root_class_names:
                parsed_types_aggregation.add("h")

            # Is this a property element (p-*, u-*, etc.) flag
            # False is default
            is_property_el = False

            # Parse plaintext p-* properties.
            p_value = None
            for prop_name in filtered_classes["p"]:
                is_property_el = True
                parsed_types_aggregation.add("p")
                prop_value = props.setdefault(prop_name, [])

                # if value has not been parsed then parse it
                if p_value is None:
                    p_value = prop_util.text(el, base_url=self.__url__)

                if root_class_names:
                    prop_value.append(
                        handle_microformat(
                            root_class_names,
                            el,
                            value_property="name",
                            simple_value=p_value,
                            backcompat_mode=backcompat_mode,
                        )
                    )
                else:
                    prop_value.append(p_value)

            # Parse URL u-* properties.
            u_value = None
            for prop_name in filtered_classes["u"]:
                is_property_el = True
                parsed_types_aggregation.add("u")
                prop_value = props.setdefault(prop_name, [])

                # if value has not been parsed then parse it
                if u_value is None:
                    u_value = prop_util.url(
                        el,
                        self.dict_class,
                        self.__img_with_alt__,
                        base_url=self.__url__,
                    )

                if root_class_names:
                    prop_value.append(
                        handle_microformat(
                            root_class_names,
                            el,
                            value_property="url",
                            simple_value=u_value,
                            backcompat_mode=backcompat_mode,
                        )
                    )
                else:
                    if isinstance(u_value, self.dict_class):
                        prop_value.append(u_value)
                    else:
                        prop_value.append(u_value)

            # Parse datetime dt-* properties.
            dt_value = None
            for prop_name in filtered_classes["dt"]:
                is_property_el = True
                parsed_types_aggregation.add("d")
                prop_value = props.setdefault(prop_name, [])

                # if value has not been parsed then parse it
                if dt_value is None:
                    dt_value, new_date = prop_util.datetime(el, self._default_date)
                    # update the default date
                    if new_date:
                        self._default_date = new_date

                if root_class_names:
                    stops_implied_name = True
                    prop_value.append(
                        handle_microformat(
                            root_class_names,
                            el,
                            simple_value=dt_value,
                            backcompat_mode=backcompat_mode,
                        )
                    )
                else:
                    if dt_value is not None:
                        prop_value.append(dt_value)

            # Parse embedded markup e-* properties.
            e_value = None
            for prop_name in filtered_classes["e"]:
                is_property_el = True
                parsed_types_aggregation.add("e")
                prop_value = props.setdefault(prop_name, [])

                # if value has not been parsed then parse it
                if e_value is None:
                    # send original element for parsing backcompat
                    if el.original is None:
                        embedded_el = el
                    else:
                        embedded_el = el.original
                    if self._preserve_doc:
                        embedded_el = copy.copy(embedded_el)
                    for el in embedded_el.find_all("template"):
                        el.extract()
                    e_value = prop_util.embedded(embedded_el, base_url=self.__url__)

                if root_class_names:
                    stops_implied_name = True
                    prop_value.append(
                        handle_microformat(
                            root_class_names,
                            el,
                            simple_value=e_value,
                            backcompat_mode=backcompat_mode,
                        )
                    )
                else:
                    prop_value.append(e_value)

            # if this is not a property element, but it is a h-* microformat,
            # add it to our list of children
            if not is_property_el and root_class_names:
                children.append(
                    handle_microformat(
                        root_class_names, el, backcompat_mode=backcompat_mode
                    )
                )
            # parse child tags, provided this isn't a microformat root-class
            if not root_class_names:
                for child in get_children(el):
                    (
                        child_properties,
                        child_microformats,
                        child_parsed_types_aggregation,
                    ) = parse_props(child)
                    for prop_name in child_properties:
                        v = props.get(prop_name, [])
                        v.extend(child_properties[prop_name])
                        props[prop_name] = v
                    children.extend(child_microformats)
                    parsed_types_aggregation.update(child_parsed_types_aggregation)
            return props, children, parsed_types_aggregation

        def parse_rels(el):
            """Parse an element for rel microformats"""
            rel_attrs = get_attr(el, "rel")
            # if rel attributes exist
            if rel_attrs is not None:
                # find the url and normalise it
                url = try_urljoin(self.__url__, el.get("href", ""))
                value_dict = self.__parsed__["rel-urls"].get(url, self.dict_class())

                # 1st one wins
                if "text" not in value_dict:
                    value_dict["text"] = el.get_text().strip()

                url_rels = value_dict.get("rels", [])
                value_dict["rels"] = url_rels

                for knownattr in ("media", "hreflang", "type", "title"):
                    x = get_attr(el, knownattr)
                    # 1st one wins
                    if x is not None and knownattr not in value_dict:
                        value_dict[knownattr] = x

                self.__parsed__["rel-urls"][url] = value_dict

                for rel_value in rel_attrs:
                    value_list = self.__parsed__["rels"].get(rel_value, [])
                    if url not in value_list:
                        value_list.append(url)
                    if rel_value not in url_rels:
                        url_rels.append(rel_value)

                    self.__parsed__["rels"][rel_value] = value_list
                if "alternate" in rel_attrs:
                    alternate_list = self.__parsed__.get("alternates", [])
                    alternate_dict = self.dict_class()
                    alternate_dict["url"] = url
                    x = " ".join([r for r in rel_attrs if not r == "alternate"])
                    if x != "":
                        alternate_dict["rel"] = x
                    alternate_dict["text"] = el.get_text().strip()
                    for knownattr in ("media", "hreflang", "type", "title"):
                        x = get_attr(el, knownattr)
                        if x is not None:
                            alternate_dict[knownattr] = x
                    alternate_list.append(alternate_dict)
                    self.__parsed__["alternates"] = alternate_list

        def parse_el(el, ctx):
            """Parse an element for microformats"""
            classes = el.get("class", [])

            # find potential microformats in root classnames h-*
            potential_microformats = class_util.root(classes)

            # if potential microformats found parse them
            if potential_microformats:
                result = handle_microformat(potential_microformats, el)
                ctx.append(result)
            else:
                # find backcompat root classnames
                potential_microformats = backcompat.root(classes)
                if potential_microformats:
                    result = handle_microformat(
                        potential_microformats, el, backcompat_mode=True
                    )
                    ctx.append(result)
                else:
                    # parse child tags
                    for child in get_children(el):
                        parse_el(child, ctx)

        ctx = []
        # start parsing at root element of the document
        parse_el(self.__doc__, ctx)
        self.__parsed__["items"] = ctx

        # parse for rel values
        for el in get_descendents(self.__doc__):
            if el.name in ("a", "area", "link") and el.has_attr("rel"):
                parse_rels(el)

        # sort the rels array in rel-urls since this should be unordered set
        for url in self.__parsed__["rel-urls"]:
            if "rels" in self.__parsed__["rel-urls"][url]:
                rels = self.__parsed__["rel-urls"][url]["rels"]
                self.__parsed__["rel-urls"][url]["rels"] = sorted(set(rels))

        # add actual parser used to debug
        # uses builder.NAME from BeautifulSoup
        if self.__html_parser__:
            self.__parsed__["debug"]["markup parser"] = self.__html_parser__
        else:
            self.__parsed__["debug"]["markup parser"] = "unknown"

    def to_dict(self, filter_by_type=None):
        """Get a dictionary version of the parsed microformat document.

        Args:
          filter_by_type (string, optional): only include top-level items of
            the given h-* type. Defaults to None.

        Returns:
            dict: representation of the parsed microformats document
        """
        if filter_by_type is None:
            return self.__parsed__
        else:
            return [x for x in self.__parsed__["items"] if filter_by_type in x["type"]]

    def to_json(self, pretty_print=False, filter_by_type=None):
        """Get a json-encoding string version of the parsed microformats document

        Args:
          pretty_print (bool, optional): Encode the json document with
            linebreaks and indents to improve readability. Defaults to False.
          filter_by_type (bool, optional): only include top-level items of
            the given h-* type

        Returns:
            string: a json-encoded string
        """

        if pretty_print:
            return json.dumps(
                self.to_dict(filter_by_type), indent=4, separators=(", ", ": ")
            )
        else:
            return json.dumps(self.to_dict(filter_by_type))
